@@ -74,150 +74,168 @@ typedef struct AutoPat
     char	    last;		// last pattern for apply_autocmds()
 } AutoPat;
 
-static struct event_name
-{
-    char	*name;	// event name
-    event_T	event;	// event number
-} event_names[] =
-{
-    {"BufAdd",		EVENT_BUFADD},
-    {"BufCreate",	EVENT_BUFADD},
-    {"BufDelete",	EVENT_BUFDELETE},
-    {"BufEnter",	EVENT_BUFENTER},
-    {"BufFilePost",	EVENT_BUFFILEPOST},
-    {"BufFilePre",	EVENT_BUFFILEPRE},
-    {"BufHidden",	EVENT_BUFHIDDEN},
-    {"BufLeave",	EVENT_BUFLEAVE},
-    {"BufNew",		EVENT_BUFNEW},
-    {"BufNewFile",	EVENT_BUFNEWFILE},
-    {"BufRead",		EVENT_BUFREADPOST},
-    {"BufReadCmd",	EVENT_BUFREADCMD},
-    {"BufReadPost",	EVENT_BUFREADPOST},
-    {"BufReadPre",	EVENT_BUFREADPRE},
-    {"BufUnload",	EVENT_BUFUNLOAD},
-    {"BufWinEnter",	EVENT_BUFWINENTER},
-    {"BufWinLeave",	EVENT_BUFWINLEAVE},
-    {"BufWipeout",	EVENT_BUFWIPEOUT},
-    {"BufWrite",	EVENT_BUFWRITEPRE},
-    {"BufWritePost",	EVENT_BUFWRITEPOST},
-    {"BufWritePre",	EVENT_BUFWRITEPRE},
-    {"BufWriteCmd",	EVENT_BUFWRITECMD},
-    {"CmdlineChanged",	EVENT_CMDLINECHANGED},
-    {"CmdlineEnter",	EVENT_CMDLINEENTER},
-    {"CmdlineLeave",	EVENT_CMDLINELEAVE},
-    {"CmdwinEnter",	EVENT_CMDWINENTER},
-    {"CmdwinLeave",	EVENT_CMDWINLEAVE},
-    {"CmdUndefined",	EVENT_CMDUNDEFINED},
-    {"ColorScheme",	EVENT_COLORSCHEME},
-    {"ColorSchemePre",	EVENT_COLORSCHEMEPRE},
-    {"CompleteChanged",	EVENT_COMPLETECHANGED},
-    {"CompleteDone",	EVENT_COMPLETEDONE},
-    {"CompleteDonePre",	EVENT_COMPLETEDONEPRE},
-    {"CursorHold",	EVENT_CURSORHOLD},
-    {"CursorHoldI",	EVENT_CURSORHOLDI},
-    {"CursorMoved",	EVENT_CURSORMOVED},
-    {"CursorMovedI",	EVENT_CURSORMOVEDI},
-    {"DiffUpdated",	EVENT_DIFFUPDATED},
-    {"DirChanged",	EVENT_DIRCHANGED},
-    {"DirChangedPre",	EVENT_DIRCHANGEDPRE},
-    {"EncodingChanged",	EVENT_ENCODINGCHANGED},
-    {"ExitPre",		EVENT_EXITPRE},
-    {"FileEncoding",	EVENT_ENCODINGCHANGED},
-    {"FileAppendPost",	EVENT_FILEAPPENDPOST},
-    {"FileAppendPre",	EVENT_FILEAPPENDPRE},
-    {"FileAppendCmd",	EVENT_FILEAPPENDCMD},
-    {"FileChangedShell",EVENT_FILECHANGEDSHELL},
-    {"FileChangedShellPost",EVENT_FILECHANGEDSHELLPOST},
-    {"FileChangedRO",	EVENT_FILECHANGEDRO},
-    {"FileReadPost",	EVENT_FILEREADPOST},
-    {"FileReadPre",	EVENT_FILEREADPRE},
-    {"FileReadCmd",	EVENT_FILEREADCMD},
-    {"FileType",	EVENT_FILETYPE},
-    {"FileWritePost",	EVENT_FILEWRITEPOST},
-    {"FileWritePre",	EVENT_FILEWRITEPRE},
-    {"FileWriteCmd",	EVENT_FILEWRITECMD},
-    {"FilterReadPost",	EVENT_FILTERREADPOST},
-    {"FilterReadPre",	EVENT_FILTERREADPRE},
-    {"FilterWritePost",	EVENT_FILTERWRITEPOST},
-    {"FilterWritePre",	EVENT_FILTERWRITEPRE},
-    {"FocusGained",	EVENT_FOCUSGAINED},
-    {"FocusLost",	EVENT_FOCUSLOST},
-    {"FuncUndefined",	EVENT_FUNCUNDEFINED},
-    {"GUIEnter",	EVENT_GUIENTER},
-    {"GUIFailed",	EVENT_GUIFAILED},
-    {"InsertChange",	EVENT_INSERTCHANGE},
-    {"InsertEnter",	EVENT_INSERTENTER},
-    {"InsertLeave",	EVENT_INSERTLEAVE},
-    {"InsertLeavePre",	EVENT_INSERTLEAVEPRE},
-    {"InsertCharPre",	EVENT_INSERTCHARPRE},
-    {"MenuPopup",	EVENT_MENUPOPUP},
-    {"ModeChanged",	EVENT_MODECHANGED},
-    {"OptionSet",	EVENT_OPTIONSET},
-    {"QuickFixCmdPost",	EVENT_QUICKFIXCMDPOST},
-    {"QuickFixCmdPre",	EVENT_QUICKFIXCMDPRE},
-    {"QuitPre",		EVENT_QUITPRE},
-    {"RemoteReply",	EVENT_REMOTEREPLY},
-    {"SafeState",	EVENT_SAFESTATE},
-    {"SafeStateAgain",	EVENT_SAFESTATEAGAIN},
-    {"SessionLoadPost",	EVENT_SESSIONLOADPOST},
-    {"ShellCmdPost",	EVENT_SHELLCMDPOST},
-    {"ShellFilterPost",	EVENT_SHELLFILTERPOST},
-    {"SigUSR1",		EVENT_SIGUSR1},
-    {"SourceCmd",	EVENT_SOURCECMD},
-    {"SourcePre",	EVENT_SOURCEPRE},
-    {"SourcePost",	EVENT_SOURCEPOST},
-    {"SpellFileMissing",EVENT_SPELLFILEMISSING},
-    {"StdinReadPost",	EVENT_STDINREADPOST},
-    {"StdinReadPre",	EVENT_STDINREADPRE},
-    {"SwapExists",	EVENT_SWAPEXISTS},
-    {"Syntax",		EVENT_SYNTAX},
-    {"TabNew",		EVENT_TABNEW},
-    {"TabClosed",	EVENT_TABCLOSED},
-    {"TabEnter",	EVENT_TABENTER},
-    {"TabLeave",	EVENT_TABLEAVE},
-    {"TermChanged",	EVENT_TERMCHANGED},
-    {"TerminalOpen",	EVENT_TERMINALOPEN},
-    {"TerminalWinOpen", EVENT_TERMINALWINOPEN},
-    {"TermResponse",	EVENT_TERMRESPONSE},
-    {"TextChanged",	EVENT_TEXTCHANGED},
-    {"TextChangedI",	EVENT_TEXTCHANGEDI},
-    {"TextChangedP",	EVENT_TEXTCHANGEDP},
-    {"TextChangedT",	EVENT_TEXTCHANGEDT},
-    {"User",		EVENT_USER},
-    {"VimEnter",	EVENT_VIMENTER},
-    {"VimLeave",	EVENT_VIMLEAVE},
-    {"VimLeavePre",	EVENT_VIMLEAVEPRE},
-    {"WinNew",		EVENT_WINNEW},
-    {"WinClosed",	EVENT_WINCLOSED},
-    {"WinEnter",	EVENT_WINENTER},
-    {"WinLeave",	EVENT_WINLEAVE},
-    {"WinResized",	EVENT_WINRESIZED},
-    {"WinScrolled",	EVENT_WINSCROLLED},
-    {"VimResized",	EVENT_VIMRESIZED},
-    {"TextYankPost",	EVENT_TEXTYANKPOST},
-    {"VimSuspend",	EVENT_VIMSUSPEND},
-    {"VimResume",	EVENT_VIMRESUME},
-    {NULL,		(event_T)0}
+//
+// special cases:
+// BufNewFile and BufRead are searched for ALOT (especially at startup)
+// so we pre-determine their index into the event_tab[] table for fast access.
+// Keep these values in sync with event_tab[]!
+#define BUFNEWFILE_INDEX 9
+#define BUFREAD_INDEX 10
+
+// must be sorted by the 'value' field because it is used by bsearch()!
+static keyvalue_T event_tab[] = {
+    KEYVALUE_ENTRY(EVENT_BUFADD, "BufAdd"),
+    KEYVALUE_ENTRY(EVENT_BUFADD, "BufCreate"),
+    KEYVALUE_ENTRY(EVENT_BUFDELETE, "BufDelete"),
+    KEYVALUE_ENTRY(EVENT_BUFENTER, "BufEnter"),
+    KEYVALUE_ENTRY(EVENT_BUFFILEPOST, "BufFilePost"),
+    KEYVALUE_ENTRY(EVENT_BUFFILEPRE, "BufFilePre"),
+    KEYVALUE_ENTRY(EVENT_BUFHIDDEN, "BufHidden"),
+    KEYVALUE_ENTRY(EVENT_BUFLEAVE, "BufLeave"),
+    KEYVALUE_ENTRY(EVENT_BUFNEW, "BufNew"),
+    KEYVALUE_ENTRY(EVENT_BUFNEWFILE, "BufNewFile"),	// BUFNEWFILE_INDEX
+    KEYVALUE_ENTRY(EVENT_BUFREADPOST, "BufRead"),	// BUFREAD_INDEX
+    KEYVALUE_ENTRY(EVENT_BUFREADCMD, "BufReadCmd"),
+    KEYVALUE_ENTRY(EVENT_BUFREADPOST, "BufReadPost"),
+    KEYVALUE_ENTRY(EVENT_BUFREADPRE, "BufReadPre"),
+    KEYVALUE_ENTRY(EVENT_BUFUNLOAD, "BufUnload"),
+    KEYVALUE_ENTRY(EVENT_BUFWINENTER, "BufWinEnter"),
+    KEYVALUE_ENTRY(EVENT_BUFWINLEAVE, "BufWinLeave"),
+    KEYVALUE_ENTRY(EVENT_BUFWIPEOUT, "BufWipeout"),
+    KEYVALUE_ENTRY(EVENT_BUFWRITEPRE, "BufWrite"),
+    KEYVALUE_ENTRY(EVENT_BUFWRITECMD, "BufWriteCmd"),
+    KEYVALUE_ENTRY(EVENT_BUFWRITEPOST, "BufWritePost"),
+    KEYVALUE_ENTRY(EVENT_BUFWRITEPRE, "BufWritePre"),
+    KEYVALUE_ENTRY(EVENT_CMDLINECHANGED, "CmdlineChanged"),
+    KEYVALUE_ENTRY(EVENT_CMDLINEENTER, "CmdlineEnter"),
+    KEYVALUE_ENTRY(EVENT_CMDLINELEAVE, "CmdlineLeave"),
+    KEYVALUE_ENTRY(EVENT_CMDUNDEFINED, "CmdUndefined"),
+    KEYVALUE_ENTRY(EVENT_CMDWINENTER, "CmdwinEnter"),
+    KEYVALUE_ENTRY(EVENT_CMDWINLEAVE, "CmdwinLeave"),
+    KEYVALUE_ENTRY(EVENT_COLORSCHEME, "ColorScheme"),
+    KEYVALUE_ENTRY(EVENT_COLORSCHEMEPRE, "ColorSchemePre"),
+    KEYVALUE_ENTRY(EVENT_COMPLETECHANGED, "CompleteChanged"),
+    KEYVALUE_ENTRY(EVENT_COMPLETEDONE, "CompleteDone"),
+    KEYVALUE_ENTRY(EVENT_COMPLETEDONEPRE, "CompleteDonePre"),
+    KEYVALUE_ENTRY(EVENT_CURSORHOLD, "CursorHold"),
+    KEYVALUE_ENTRY(EVENT_CURSORHOLDI, "CursorHoldI"),
+    KEYVALUE_ENTRY(EVENT_CURSORMOVED, "CursorMoved"),
+    KEYVALUE_ENTRY(EVENT_CURSORMOVEDC, "CursorMovedC"),
+    KEYVALUE_ENTRY(EVENT_CURSORMOVEDI, "CursorMovedI"),
+    KEYVALUE_ENTRY(EVENT_DIFFUPDATED, "DiffUpdated"),
+    KEYVALUE_ENTRY(EVENT_DIRCHANGED, "DirChanged"),
+    KEYVALUE_ENTRY(EVENT_DIRCHANGEDPRE, "DirChangedPre"),
+    KEYVALUE_ENTRY(EVENT_ENCODINGCHANGED, "EncodingChanged"),
+    KEYVALUE_ENTRY(EVENT_EXITPRE, "ExitPre"),
+    KEYVALUE_ENTRY(EVENT_FILEAPPENDCMD, "FileAppendCmd"),
+    KEYVALUE_ENTRY(EVENT_FILEAPPENDPOST, "FileAppendPost"),
+    KEYVALUE_ENTRY(EVENT_FILEAPPENDPRE, "FileAppendPre"),
+    KEYVALUE_ENTRY(EVENT_FILECHANGEDRO, "FileChangedRO"),
+    KEYVALUE_ENTRY(EVENT_FILECHANGEDSHELL, "FileChangedShell"),
+    KEYVALUE_ENTRY(EVENT_FILECHANGEDSHELLPOST, "FileChangedShellPost"),
+    KEYVALUE_ENTRY(EVENT_ENCODINGCHANGED, "FileEncoding"),
+    KEYVALUE_ENTRY(EVENT_FILEREADCMD, "FileReadCmd"),
+    KEYVALUE_ENTRY(EVENT_FILEREADPOST, "FileReadPost"),
+    KEYVALUE_ENTRY(EVENT_FILEREADPRE, "FileReadPre"),
+    KEYVALUE_ENTRY(EVENT_FILETYPE, "FileType"),
+    KEYVALUE_ENTRY(EVENT_FILEWRITECMD, "FileWriteCmd"),
+    KEYVALUE_ENTRY(EVENT_FILEWRITEPOST, "FileWritePost"),
+    KEYVALUE_ENTRY(EVENT_FILEWRITEPRE, "FileWritePre"),
+    KEYVALUE_ENTRY(EVENT_FILTERREADPOST, "FilterReadPost"),
+    KEYVALUE_ENTRY(EVENT_FILTERREADPRE, "FilterReadPre"),
+    KEYVALUE_ENTRY(EVENT_FILTERWRITEPOST, "FilterWritePost"),
+    KEYVALUE_ENTRY(EVENT_FILTERWRITEPRE, "FilterWritePre"),
+    KEYVALUE_ENTRY(EVENT_FOCUSGAINED, "FocusGained"),
+    KEYVALUE_ENTRY(EVENT_FOCUSLOST, "FocusLost"),
+    KEYVALUE_ENTRY(EVENT_FUNCUNDEFINED, "FuncUndefined"),
+    KEYVALUE_ENTRY(EVENT_GUIENTER, "GUIEnter"),
+    KEYVALUE_ENTRY(EVENT_GUIFAILED, "GUIFailed"),
+    KEYVALUE_ENTRY(EVENT_INSERTCHANGE, "InsertChange"),
+    KEYVALUE_ENTRY(EVENT_INSERTCHARPRE, "InsertCharPre"),
+    KEYVALUE_ENTRY(EVENT_INSERTENTER, "InsertEnter"),
+    KEYVALUE_ENTRY(EVENT_INSERTLEAVE, "InsertLeave"),
+    KEYVALUE_ENTRY(EVENT_INSERTLEAVEPRE, "InsertLeavePre"),
+    KEYVALUE_ENTRY(EVENT_KEYINPUTPRE, "KeyInputPre"),
+    KEYVALUE_ENTRY(EVENT_MENUPOPUP, "MenuPopup"),
+    KEYVALUE_ENTRY(EVENT_MODECHANGED, "ModeChanged"),
+    KEYVALUE_ENTRY(EVENT_OPTIONSET, "OptionSet"),
+    KEYVALUE_ENTRY(EVENT_QUICKFIXCMDPOST, "QuickFixCmdPost"),
+    KEYVALUE_ENTRY(EVENT_QUICKFIXCMDPRE, "QuickFixCmdPre"),
+    KEYVALUE_ENTRY(EVENT_QUITPRE, "QuitPre"),
+    KEYVALUE_ENTRY(EVENT_REMOTEREPLY, "RemoteReply"),
+    KEYVALUE_ENTRY(EVENT_SAFESTATE, "SafeState"),
+    KEYVALUE_ENTRY(EVENT_SAFESTATEAGAIN, "SafeStateAgain"),
+    KEYVALUE_ENTRY(EVENT_SESSIONLOADPOST, "SessionLoadPost"),
+    KEYVALUE_ENTRY(EVENT_SESSIONWRITEPOST, "SessionWritePost"),
+    KEYVALUE_ENTRY(EVENT_SHELLCMDPOST, "ShellCmdPost"),
+    KEYVALUE_ENTRY(EVENT_SHELLFILTERPOST, "ShellFilterPost"),
+    KEYVALUE_ENTRY(EVENT_SIGUSR1, "SigUSR1"),
+    KEYVALUE_ENTRY(EVENT_SOURCECMD, "SourceCmd"),
+    KEYVALUE_ENTRY(EVENT_SOURCEPOST, "SourcePost"),
+    KEYVALUE_ENTRY(EVENT_SOURCEPRE, "SourcePre"),
+    KEYVALUE_ENTRY(EVENT_SPELLFILEMISSING, "SpellFileMissing"),
+    KEYVALUE_ENTRY(EVENT_STDINREADPOST, "StdinReadPost"),
+    KEYVALUE_ENTRY(EVENT_STDINREADPRE, "StdinReadPre"),
+    KEYVALUE_ENTRY(EVENT_SWAPEXISTS, "SwapExists"),
+    KEYVALUE_ENTRY(EVENT_SYNTAX, "Syntax"),
+    KEYVALUE_ENTRY(EVENT_TABCLOSED, "TabClosed"),
+    KEYVALUE_ENTRY(EVENT_TABENTER, "TabEnter"),
+    KEYVALUE_ENTRY(EVENT_TABLEAVE, "TabLeave"),
+    KEYVALUE_ENTRY(EVENT_TABNEW, "TabNew"),
+    KEYVALUE_ENTRY(EVENT_TERMCHANGED, "TermChanged"),
+    KEYVALUE_ENTRY(EVENT_TERMINALOPEN, "TerminalOpen"),
+    KEYVALUE_ENTRY(EVENT_TERMINALWINOPEN, "TerminalWinOpen"),
+    KEYVALUE_ENTRY(EVENT_TERMRESPONSE, "TermResponse"),
+    KEYVALUE_ENTRY(EVENT_TERMRESPONSEALL, "TermResponseAll"),
+    KEYVALUE_ENTRY(EVENT_TEXTCHANGED, "TextChanged"),
+    KEYVALUE_ENTRY(EVENT_TEXTCHANGEDI, "TextChangedI"),
+    KEYVALUE_ENTRY(EVENT_TEXTCHANGEDP, "TextChangedP"),
+    KEYVALUE_ENTRY(EVENT_TEXTCHANGEDT, "TextChangedT"),
+    KEYVALUE_ENTRY(EVENT_TEXTYANKPOST, "TextYankPost"),
+    KEYVALUE_ENTRY(EVENT_USER, "User"),
+    KEYVALUE_ENTRY(EVENT_VIMENTER, "VimEnter"),
+    KEYVALUE_ENTRY(EVENT_VIMLEAVE, "VimLeave"),
+    KEYVALUE_ENTRY(EVENT_VIMLEAVEPRE, "VimLeavePre"),
+    KEYVALUE_ENTRY(EVENT_VIMRESIZED, "VimResized"),
+    KEYVALUE_ENTRY(EVENT_VIMRESUME, "VimResume"),
+    KEYVALUE_ENTRY(EVENT_VIMSUSPEND, "VimSuspend"),
+    KEYVALUE_ENTRY(EVENT_WINCLOSED, "WinClosed"),
+    KEYVALUE_ENTRY(EVENT_WINENTER, "WinEnter"),
+    KEYVALUE_ENTRY(EVENT_WINLEAVE, "WinLeave"),
+    KEYVALUE_ENTRY(EVENT_WINNEW, "WinNew"),
+    KEYVALUE_ENTRY(EVENT_WINNEWPRE, "WinNewPre"),
+    KEYVALUE_ENTRY(EVENT_WINRESIZED, "WinResized"),
+    KEYVALUE_ENTRY(EVENT_WINSCROLLED, "WinScrolled")
 };
 
-static AutoPat *first_autopat[NUM_EVENTS] =
-{
+static AutoPat *first_autopat[NUM_EVENTS] = {
     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
-    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
+    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+    NULL, NULL, NULL, NULL, NULL, NULL
 };
 
-static AutoPat *last_autopat[NUM_EVENTS] =
-{
+static AutoPat *last_autopat[NUM_EVENTS] = {
     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
-    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
+    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+    NULL, NULL, NULL, NULL, NULL, NULL
 };
 
 #define AUGROUP_DEFAULT    (-1)	    // default autocmd group
@@ -263,6 +281,7 @@ static int current_augroup = AUGROUP_DEFAULT;
 
 static int au_need_clean = FALSE;   // need to delete marked patterns
 
+static event_T event_name2nr(char_u *start, char_u **end);
 static char_u *event_nr2name(event_T event);
 static int au_get_grouparg(char_u **argp);
 static int do_autocmd_event(event_T event, char_u *pat, int once, int nested, char_u *cmd, int forceit, int group, int flags);
@@ -678,24 +697,35 @@ is_aucmd_win(win_T *win)
 event_name2nr(char_u *start, char_u **end)
 {
     char_u	*p;
-    int		i;
-    int		len;
+    keyvalue_T target;
+    keyvalue_T *entry;
+    static keyvalue_T *bufnewfile = &event_tab[BUFNEWFILE_INDEX];
+    static keyvalue_T *bufread = &event_tab[BUFREAD_INDEX];
 
     // the event name ends with end of line, '|', a blank or a comma
     for (p = start; *p && !VIM_ISWHITE(*p) && *p != ',' && *p != '|'; ++p)
 	;
-    for (i = 0; event_names[i].name != NULL; ++i)
-    {
-	len = (int)STRLEN(event_names[i].name);
-	if (len == p - start && STRNICMP(event_names[i].name, start, len) == 0)
-	    break;
-    }
+
+    target.key = 0;
+    target.value.string = start;
+    target.value.length = (size_t)(p - start);
+
+    // special cases:
+    // BufNewFile and BufRead are searched for ALOT (especially at startup)
+    // so we check for them first.
+    if (cmp_keyvalue_value_ni(&target, bufnewfile) == 0)
+	entry = bufnewfile;
+    else
+    if (cmp_keyvalue_value_ni(&target, bufread) == 0)
+	entry = bufread;
+    else
+	entry = (keyvalue_T *)bsearch(&target, &event_tab, ARRAY_LENGTH(event_tab), sizeof(event_tab[0]), cmp_keyvalue_value_ni);
+
     if (*p == ',')
 	++p;
     *end = p;
-    if (event_names[i].name == NULL)
-	return NUM_EVENTS;
-    return event_names[i].event;
+
+    return (entry == NULL) ? NUM_EVENTS : (event_T)entry->key;
 }
 
 /*
@@ -705,11 +735,53 @@ event_name2nr(char_u *start, char_u **end)
 event_nr2name(event_T event)
 {
     int	    i;
+#define CACHE_SIZE 12
+    static int cache_tab[CACHE_SIZE];
+    static int cache_last_index = -1;
 
-    for (i = 0; event_names[i].name != NULL; ++i)
-	if (event_names[i].event == event)
-	    return (char_u *)event_names[i].name;
-    return (char_u *)"Unknown";
+    if (cache_last_index < 0)
+    {
+	for (i = 0; i < (int)ARRAY_LENGTH(cache_tab); ++i)
+	    cache_tab[i] = -1;
+	cache_last_index = ARRAY_LENGTH(cache_tab) - 1;
+    }
+
+    // first look in the cache
+    // the cache is circular. to search it we start at the most recent entry
+    // and go backwards wrapping around when we get to index 0.
+    for (i = cache_last_index; cache_tab[i] >= 0; )
+    {
+	if ((event_T)event_tab[cache_tab[i]].key == event)
+	    return event_tab[cache_tab[i]].value.string;
+
+	if (i == 0)
+	    i = ARRAY_LENGTH(cache_tab) - 1;
+	else
+	    --i;
+
+	// are we back at the start?
+	if (i == cache_last_index)
+	    break;
+    }
+
+    // look in the event table itself
+    for (i = 0; i < (int)ARRAY_LENGTH(event_tab); ++i)
+    {
+	if ((event_T)event_tab[i].key == event)
+	{
+	    // store the found entry in the next position in the cache,
+	    // wrapping around when we get to the maximum index.
+	    if (cache_last_index == ARRAY_LENGTH(cache_tab) - 1)
+		cache_last_index = 0;
+	    else
+		++cache_last_index;
+	    cache_tab[cache_last_index] = i;
+	    break;
+	}
+    }
+
+    return (i == (int)ARRAY_LENGTH(event_tab)) ? (char_u *)"Unknown" :
+	event_tab[i].value.string;
 }
 
 /*
@@ -803,12 +875,14 @@ au_event_disable(char *what)
 {
     char_u	*new_ei;
     char_u	*save_ei;
+    size_t	p_ei_len;
 
-    save_ei = vim_strsave(p_ei);
+    p_ei_len = STRLEN(p_ei);
+    save_ei = vim_strnsave(p_ei, p_ei_len);
     if (save_ei == NULL)
 	return NULL;
 
-    new_ei = vim_strnsave(p_ei, STRLEN(p_ei) + STRLEN(what));
+    new_ei = vim_strnsave(p_ei, p_ei_len + STRLEN(what));
     if (new_ei == NULL)
     {
 	vim_free(save_ei);
@@ -1564,9 +1638,12 @@ aucmd_prepbuf(
     }
 
     aco->save_curwin_id = curwin->w_id;
-    aco->save_curbuf = curbuf;
     aco->save_prevwin_id = prevwin == NULL ? 0 : prevwin->w_id;
     aco->save_State = State;
+#ifdef FEAT_JOB_CHANNEL
+    if (bt_prompt(curbuf))
+	aco->save_prompt_insert = curbuf->b_prompt_insert;
+#endif
 
     if (win != NULL)
     {
@@ -1586,6 +1663,11 @@ aucmd_prepbuf(
 
 	win_init_popup_win(auc_win, buf);
 
+	// Make sure tp_localdir and globaldir are NULL to avoid a
+	// chdir() in win_enter_ext().
+	// win_init_popup_win() has already set w_localdir to NULL.
+	aco->tp_localdir = curtab->tp_localdir;
+	curtab->tp_localdir = NULL;
 	aco->globaldir = globaldir;
 	globaldir = NULL;
 
@@ -1602,7 +1684,7 @@ aucmd_prepbuf(
 	p_acd = FALSE;
 #endif
 
-	(void)win_split_ins(0, WSP_TOP, auc_win, 0);
+	(void)win_split_ins(0, WSP_TOP | WSP_FORCE_ROOM, auc_win, 0, NULL);
 	(void)win_comp_pos();   // recompute window positions
 	p_ea = save_ea;
 #ifdef FEAT_AUTOCHDIR
@@ -1635,7 +1717,6 @@ aucmd_restbuf(
     {
 	win_T *awp = aucmd_win[aco->use_aucmd_win_idx].auc_win;
 
-	--curbuf->b_nwindows;
 	// Find "awp", it can't be closed, but it may be in another tab
 	// page. Do not trigger autocommands here.
 	block_autocmds();
@@ -1656,8 +1737,8 @@ aucmd_restbuf(
 	    }
 	}
 win_found:
+	--curbuf->b_nwindows;
 #ifdef FEAT_JOB_CHANNEL
-	;
 	int save_stop_insert_mode = stop_insert_mode;
 	// May need to stop Insert mode if we were in a prompt buffer.
 	leaving_window(curwin);
@@ -1666,7 +1747,7 @@ win_found:
 	    stop_insert_mode = save_stop_insert_mode;
 #endif
 	// Remove the window and frame from the tree of frames.
-	(void)winframe_remove(curwin, &dummy, NULL);
+	(void)winframe_remove(curwin, &dummy, NULL, NULL);
 	win_remove(curwin, NULL);
 
 	// The window is marked as not used, but it is not freed, it can be
@@ -1692,12 +1773,20 @@ win_found:
 #ifdef FEAT_JOB_CHANNEL
 	// May need to restore insert mode for a prompt buffer.
 	entering_window(curwin);
+	if (bt_prompt(curbuf))
+	    curbuf->b_prompt_insert = aco->save_prompt_insert;
 #endif
 	prevwin = win_find_by_id(aco->save_prevwin_id);
 #ifdef FEAT_EVAL
 	vars_clear(&awp->w_vars->dv_hashtab);  // free all w: variables
 	hash_init(&awp->w_vars->dv_hashtab);   // re-use the hashtab
 #endif
+	// If :lcd has been used in the autocommand window, correct current
+	// directory before restoring tp_localdir and globaldir.
+	if (awp->w_localdir != NULL)
+	    win_fix_current_dir();
+	vim_free(curtab->tp_localdir);
+	curtab->tp_localdir = aco->tp_localdir;
 	vim_free(globaldir);
 	globaldir = aco->globaldir;
 
@@ -1934,6 +2023,15 @@ has_insertcharpre(void)
     return (first_autopat[(int)EVENT_INSERTCHARPRE] != NULL);
 }
 
+/*
+ * Return TRUE when there is an KeyInputPre autocommand defined.
+ */
+    int
+has_keyinputpre(void)
+{
+    return (first_autopat[(int)EVENT_KEYINPUTPRE] != NULL);
+}
+
 /*
  * Return TRUE when there is an CmdUndefined autocommand defined.
  */
@@ -2101,7 +2199,8 @@ apply_autocmds_group(
     {
 	if (event == EVENT_COLORSCHEME || event == EVENT_COLORSCHEMEPRE
 						 || event == EVENT_OPTIONSET
-						 || event == EVENT_MODECHANGED)
+						 || event == EVENT_MODECHANGED
+						 || event == EVENT_TERMRESPONSEALL)
 	    autocmd_fname = NULL;
 	else if (fname != NULL && !ends_excmd(*fname))
 	    autocmd_fname = fname;
@@ -2163,10 +2262,12 @@ apply_autocmds_group(
 		|| event == EVENT_CMDLINECHANGED
 		|| event == EVENT_CMDLINEENTER
 		|| event == EVENT_CMDLINELEAVE
+		|| event == EVENT_CURSORMOVEDC
 		|| event == EVENT_CMDWINENTER
 		|| event == EVENT_CMDWINLEAVE
 		|| event == EVENT_CMDUNDEFINED
 		|| event == EVENT_FUNCUNDEFINED
+		|| event == EVENT_KEYINPUTPRE
 		|| event == EVENT_REMOTEREPLY
 		|| event == EVENT_SPELLFILEMISSING
 		|| event == EVENT_QUICKFIXCMDPRE
@@ -2181,7 +2282,8 @@ apply_autocmds_group(
 		|| event == EVENT_USER
 		|| event == EVENT_WINCLOSED
 		|| event == EVENT_WINRESIZED
-		|| event == EVENT_WINSCROLLED)
+		|| event == EVENT_WINSCROLLED
+		|| event == EVENT_TERMRESPONSEALL)
 	{
 	    fname = vim_strsave(fname);
 	    autocmd_fname_full = TRUE; // don't expand it later
@@ -2249,7 +2351,7 @@ apply_autocmds_group(
 	    saveRedobuff(&save_redo);
 	    did_save_redobuff = TRUE;
 	}
-	did_filetype = keep_filetype;
+	curbuf->b_did_filetype = curbuf->b_keep_filetype;
     }
 
     /*
@@ -2261,7 +2363,7 @@ apply_autocmds_group(
 
     // Remember that FileType was triggered.  Used for did_filetype().
     if (event == EVENT_FILETYPE)
-	did_filetype = TRUE;
+	curbuf->b_did_filetype = TRUE;
 
     tail = gettail(fname);
 
@@ -2370,7 +2472,7 @@ apply_autocmds_group(
 	restore_search_patterns();
 	if (did_save_redobuff)
 	    restoreRedobuff(&save_redo);
-	did_filetype = FALSE;
+	curbuf->b_did_filetype = FALSE;
 	while (au_pending_free_buf != NULL)
 	{
 	    buf_T *b = au_pending_free_buf->b_next;
@@ -2412,13 +2514,18 @@ BYPASS_AU:
 	aubuflocal_remove(buf);
 
     if (retval == OK && event == EVENT_FILETYPE)
-	au_did_filetype = TRUE;
+	curbuf->b_au_did_filetype = TRUE;
 
     return retval;
 }
 
 # ifdef FEAT_EVAL
 static char_u	*old_termresponse = NULL;
+static char_u	*old_termu7resp = NULL;
+static char_u	*old_termblinkresp = NULL;
+static char_u	*old_termrbgresp = NULL;
+static char_u	*old_termrfgresp = NULL;
+static char_u	*old_termstyleresp = NULL;
 # endif
 
 /*
@@ -2431,7 +2538,14 @@ block_autocmds(void)
 # ifdef FEAT_EVAL
     // Remember the value of v:termresponse.
     if (autocmd_blocked == 0)
+    {
 	old_termresponse = get_vim_var_str(VV_TERMRESPONSE);
+	old_termu7resp = get_vim_var_str(VV_TERMU7RESP);
+	old_termblinkresp = get_vim_var_str(VV_TERMBLINKRESP);
+	old_termrbgresp = get_vim_var_str(VV_TERMRBGRESP);
+	old_termrfgresp = get_vim_var_str(VV_TERMRFGRESP);
+	old_termstyleresp = get_vim_var_str(VV_TERMSTYLERESP);
+    }
 # endif
     ++autocmd_blocked;
 }
@@ -2442,12 +2556,37 @@ unblock_autocmds(void)
     --autocmd_blocked;
 
 # ifdef FEAT_EVAL
-    // When v:termresponse was set while autocommands were blocked, trigger
-    // the autocommands now.  Esp. useful when executing a shell command
-    // during startup (vimdiff).
-    if (autocmd_blocked == 0
-		      && get_vim_var_str(VV_TERMRESPONSE) != old_termresponse)
-	apply_autocmds(EVENT_TERMRESPONSE, NULL, NULL, FALSE, curbuf);
+    // When v:termresponse, etc, were set while autocommands were blocked,
+    // trigger the autocommands now.  Esp. useful when executing a shell
+    // command during startup (vimdiff).
+    if (autocmd_blocked == 0)
+    {
+	if (get_vim_var_str(VV_TERMRESPONSE) != old_termresponse)
+	{
+	    apply_autocmds(EVENT_TERMRESPONSE, NULL, NULL, FALSE, curbuf);
+	    apply_autocmds(EVENT_TERMRESPONSEALL, (char_u *)"version", NULL, FALSE, curbuf);
+	}
+	if (get_vim_var_str(VV_TERMU7RESP) != old_termu7resp)
+	{
+	    apply_autocmds(EVENT_TERMRESPONSEALL, (char_u *)"ambiguouswidth", NULL, FALSE, curbuf);
+	}
+	if (get_vim_var_str(VV_TERMBLINKRESP) != old_termblinkresp)
+	{
+	    apply_autocmds(EVENT_TERMRESPONSEALL, (char_u *)"cursorblink", NULL, FALSE, curbuf);
+	}
+	if (get_vim_var_str(VV_TERMRBGRESP) != old_termrbgresp)
+	{
+	    apply_autocmds(EVENT_TERMRESPONSEALL, (char_u *)"background", NULL, FALSE, curbuf);
+	}
+	if (get_vim_var_str(VV_TERMRFGRESP) != old_termrfgresp)
+	{
+	    apply_autocmds(EVENT_TERMRESPONSEALL, (char_u *)"foreground", NULL, FALSE, curbuf);
+	}
+	if (get_vim_var_str(VV_TERMSTYLERESP) != old_termstyleresp)
+	{
+	    apply_autocmds(EVENT_TERMRESPONSEALL, (char_u *)"cursorshape", NULL, FALSE, curbuf);
+	}
+    }
 # endif
 }
 
@@ -2728,6 +2867,8 @@ set_context_in_autocmd(
     char_u *
 get_event_name(expand_T *xp UNUSED, int idx)
 {
+    int i;
+
     if (idx < augroups.ga_len)		// First list group names, if wanted
     {
 	if (!include_groups || AUGROUP_NAME(idx) == NULL
@@ -2735,7 +2876,25 @@ get_event_name(expand_T *xp UNUSED, int idx)
 	    return (char_u *)"";	// skip deleted entries
 	return AUGROUP_NAME(idx);	// return a name
     }
-    return (char_u *)event_names[idx - augroups.ga_len].name;
+
+    i = idx - augroups.ga_len;
+    if (i < 0 || i >= (int)ARRAY_LENGTH(event_tab))
+	return NULL;
+
+    return event_tab[i].value.string;
+}
+
+/*
+ * Function given to ExpandGeneric() to obtain the list of event names. Don't
+ * include groups.
+ */
+    char_u *
+get_event_name_no_group(expand_T *xp UNUSED, int idx)
+{
+    if (idx < 0 || idx >= (int)ARRAY_LENGTH(event_tab))
+	return NULL;
+
+    return event_tab[idx].value.string;
 }
 
 
@@ -3195,8 +3354,6 @@ f_autocmd_get(typval_T *argvars, typval_T *rettv)
 	// return only the autocmds for the specified event
 	if (dict_has_key(argvars[0].vval.v_dict, "event"))
 	{
-	    int		i;
-
 	    name = dict_get_string(argvars[0].vval.v_dict, "event", TRUE);
 	    if (name == NULL)
 		return;
@@ -3205,16 +3362,22 @@ f_autocmd_get(typval_T *argvars, typval_T *rettv)
 		event_arg = NUM_EVENTS;
 	    else
 	    {
-		for (i = 0; event_names[i].name != NULL; i++)
-		    if (STRICMP(event_names[i].name, name) == 0)
-			break;
-		if (event_names[i].name == NULL)
+		keyvalue_T target;
+		keyvalue_T *entry;
+
+		target.key = 0;
+		target.value.string = name;
+		target.value.length = STRLEN(target.value.string);
+		entry = (keyvalue_T *)bsearch(&target, &event_tab,
+			ARRAY_LENGTH(event_tab), sizeof(event_tab[0]),
+			cmp_keyvalue_value_ni);
+		if (entry == NULL)
 		{
 		    semsg(_(e_no_such_event_str), name);
 		    vim_free(name);
 		    return;
 		}
-		event_arg = event_names[i].event;
+		event_arg = (event_T)entry->key;
 	    }
 	    vim_free(name);
 	}
@@ -3244,6 +3407,9 @@ f_autocmd_get(typval_T *argvars, typval_T *rettv)
 	{
 	    char_u	*group_name;
 
+	    if (ap->pat == NULL)		// pattern has been removed
+		continue;
+
 	    if (group != AUGROUP_ALL && group != ap->group)
 		continue;
 
@@ -3259,7 +3425,10 @@ f_autocmd_get(typval_T *argvars, typval_T *rettv)
 		event_dict = dict_alloc();
 		if (event_dict == NULL
 			|| list_append_dict(event_list, event_dict) == FAIL)
+		{
+		    vim_free(pat);
 		    return;
+		}
 
 		if (dict_add_string(event_dict, "event", event_name) == FAIL
 			|| dict_add_string(event_dict, "group",
@@ -3274,7 +3443,10 @@ f_autocmd_get(typval_T *argvars, typval_T *rettv)
 			|| dict_add_bool(event_dict, "once", ac->once) == FAIL
 			|| dict_add_bool(event_dict, "nested",
 							   ac->nested) == FAIL)
+		{
+		    vim_free(pat);
 		    return;
+		}
 	    }
 	}
     }
