@@ -325,7 +325,7 @@ im_preedit_window_open(void)
 #if GTK_CHECK_VERSION(3,16,0)
     {
 	GtkStyleContext * const context
-				  = gtk_widget_get_style_context(gui.drawarea);
+				  = gtk_widget_get_style_context(preedit_label);
 	GtkCssProvider * const provider = gtk_css_provider_new();
 	gchar		   *css = NULL;
 	const char * const fontname
@@ -349,7 +349,7 @@ im_preedit_window_open(void)
 	    fontsize_propval = g_strdup_printf("inherit");
 
 	css = g_strdup_printf(
-		"widget#vim-gui-preedit-area {\n"
+		"#vim-gui-preedit-area {\n"
 		"  font-family: %s,monospace;\n"
 		"  font-size: %s;\n"
 		"  color: #%.2lx%.2lx%.2lx;\n"
@@ -1063,6 +1063,9 @@ xim_reset(void)
     int
 xim_queue_key_press_event(GdkEventKey *event, int down)
 {
+#ifdef FEAT_GUI_GTK
+    if (event->state & GDK_SUPER_MASK) return FALSE;
+#endif
     if (down)
     {
 	// Workaround GTK2 XIM 'feature' that always converts keypad keys to
@@ -1471,7 +1474,7 @@ xim_real_init(Window x11_window, Display *x11_display)
 		break;
 	    if ((ns = end = strchr(s, ',')) == NULL)
 		end = s + strlen(s);
-	    while (isspace(((char_u *)end)[-1]))
+	    while (SAFE_isspace(end[-1]))
 		end--;
 	    *end = NUL;
 
@@ -1533,7 +1536,7 @@ xim_real_init(Window x11_window, Display *x11_display)
     strcpy(tmp, gui.rsrc_preedit_type_name);
     for (s = tmp; s && !found; )
     {
-	while (*s && isspace((unsigned char)*s))
+	while (*s && SAFE_isspace(*s))
 	    s++;
 	if (!*s)
 	    break;
@@ -1541,7 +1544,7 @@ xim_real_init(Window x11_window, Display *x11_display)
 	    ns++;
 	else
 	    end = s + strlen(s);
-	while (isspace((unsigned char)*end))
+	while (SAFE_isspace(*end))
 	    end--;
 	*end = '\0';
 
