@@ -489,8 +489,3 @@ int mch_rename(const char *src, const char *dest);
 // We have three kinds of ACL support.
 #define HAVE_ACL (HAVE_POSIX_ACL || HAVE_SOLARIS_ACL || HAVE_AIX_ACL)
 
-struct cellsize {
-    unsigned int cs_xpixel;
-    unsigned int cs_ypixel;
-};
-
